import json
import sys
import pytest

from addok import ds
from addok.autocomplete import create_edge_ngrams, index_edge_ngrams
from addok.batch import process_documents
from addok.db import DB


def index_document(doc):
    process_documents(json.dumps(doc))


def deindex_document(_id):
    process_documents(json.dumps({"_id": _id, "_action": "delete"}))


def count_keys():
    """Helper method to return the number of keys in the test database."""
    try:
        return DB.info()["db14"]["keys"]
    except KeyError:
        return 0


def test_index_edge_ngrams():
    before = count_keys()
    index_edge_ngrams(DB, "street")
    after = count_keys()
    assert after - before == 3
    assert DB.smembers("n|str") == set([b"street"])
    assert DB.smembers("n|stre") == set([b"street"])
    assert DB.smembers("n|stree") == set([b"street"])


DOC = {
    "id": "xxxx",
    "_id": "yyyy",
    "type": "street",
    "name": "rue des Lilas",
    "city": "Andrésy",
    "lat": "48.32545",
    "lon": "2.2565",
    "housenumbers": {"1": {"lat": "48.325451", "lon": "2.25651"}},
}


def test_index_document():
    index_document(DOC.copy())
    assert ds._DB.exists("d|yyyy")
    assert ds._DB.type("d|yyyy") == b"string"
    assert DB.exists("w|rue")
    assert b"d|yyyy" in DB.zrange("w|rue", 0, -1)
    assert DB.exists("w|des")
    assert DB.exists("w|lilas")
    assert DB.exists("w|andresy")
    assert DB.exists("p|rue")
    assert DB.exists("p|des")
    assert DB.exists("p|lilas")
    assert DB.exists("p|andresy")
    assert b"lilas" in DB.smembers("p|andresy")
    assert b"andresy" in DB.smembers("p|lilas")
    assert DB.exists("g|u09dgm7")
    assert b"d|yyyy" in DB.smembers("g|u09dgm7")
    assert DB.exists("n|lil")
    assert DB.exists("n|lila")
    assert DB.exists("n|and")
    assert b"andresy" in DB.smembers("n|and")
    assert DB.exists("n|andr")
    assert b"andresy" in DB.smembers("n|andr")
    assert DB.exists("n|andre")
    assert b"andresy" in DB.smembers("n|andre")
    assert DB.exists("n|andres")
    assert b"andresy" in DB.smembers("n|andres")
    assert b"lilas" in DB.smembers("n|lil")
    assert DB.exists("f|type|street")
    assert b"d|yyyy" in DB.smembers("f|type|street")
    assert DB.exists("f|type|housenumber")
    assert b"d|yyyy" in DB.smembers("f|type|housenumber")
    assert len(DB.keys()) == 17
    assert len(ds._DB.keys()) == 1


def test_index_document_without_explicit_id():
    doc = DOC.copy()
    del doc["_id"]
    index_document(doc)
    assert ds._DB.exists("d|jR")
    assert ds._DB.type("d|jR") == b"string"
    assert DB.exists("w|rue")


def test_deindex_document_should_deindex():
    index_document(DOC.copy())
    deindex_document(DOC["_id"])
    assert not ds._DB.exists("d|xxxx")
    assert not ds._DB.exists("d|yyyy")
    assert not DB.exists("w|de")
    assert not DB.exists("w|lilas")
    assert not DB.exists("w|1")  # Housenumber.
    assert not DB.exists("p|rue")
    assert not DB.exists("p|des")
    assert not DB.exists("p|lilas")
    assert not DB.exists("p|1")
    assert not DB.exists("g|u09dgm7")
    assert not DB.exists("n|lil")
    assert not DB.exists("n|and")
    assert not DB.exists("n|andr")
    assert not DB.exists("n|andre")
    assert not DB.exists("n|andres")
    assert not DB.exists("f|type|street")
    assert len(DB.keys()) == 0
    assert len(ds._DB.keys()) == 0


def test_deindex_document_should_not_affect_other_docs():
    DOC2 = {
        "id": "xxxx2",
        "_id": "yyyy2",
        "type": "street",
        "name": "rue des Lilas",
        "city": "Paris",
        "lat": "49.32545",
        "lon": "4.2565",
        "housenumbers": {
            "1": {"lat": "48.325451", "lon": "2.25651"}  # Same geohash as DOC.
        },
    }
    DOC1 = json.loads(json.dumps(DOC))  # deepcopy.
    # Add new housenumber so we can check it's deindexed.
    DOC1["housenumbers"]["2"] = {"lat": "48.325459", "lon": "2.25659"}
    index_document(DOC1)
    index_document(DOC2)
    deindex_document(DOC1["_id"])
    assert not ds._DB.exists("d|yyyy")
    assert b"d|yyyy" not in DB.zrange("w|rue", 0, -1)
    assert b"d|yyyy" not in DB.zrange("w|des", 0, -1)
    assert b"d|yyyy" not in DB.zrange("w|lilas", 0, -1)
    assert DB.exists("g|u09dgm7")
    assert b"d|yyyy" not in DB.smembers("g|u09dgm7")
    assert DB.exists("w|des")
    assert DB.exists("w|lilas")
    assert DB.exists("p|rue")
    assert b"d|yyyy2" in DB.zrange("w|rue", 0, -1)
    assert b"d|yyyy2" in DB.zrange("w|des", 0, -1)
    assert b"d|yyyy2" in DB.zrange("w|lilas", 0, -1)
    assert b"d|yyyy2" in DB.smembers("g|u09dgm7")
    assert b"d|yyyy2" in DB.smembers("g|u0g08g7")
    assert DB.exists("p|des")
    assert DB.exists("p|lilas")
    assert not DB.exists("n|and")
    assert not DB.exists("n|andr")
    assert not DB.exists("n|andre")
    assert not DB.exists("n|andres")
    assert DB.exists("n|par")
    assert DB.exists("n|pari")
    assert DB.exists("n|lil")
    assert DB.exists("n|lila")
    assert b"lilas" in DB.smembers("n|lil")
    assert b"lilas" in DB.smembers("n|lila")
    assert DB.exists("f|type|street")
    assert b"d|yyyy2" in DB.smembers("f|type|street")
    assert DB.exists("f|type|housenumber")
    assert b"d|yyyy2" in DB.smembers("f|type|housenumber")
    assert len(DB.keys()) == 16
    assert len(ds._DB.keys()) == 1


def test_allow_list_values():
    doc = {
        "id": "xxxx",
        "_id": "yyyy",
        "type": "street",
        "name": ["Vernou-la-Celle-sur-Seine", "Vernou"],
        "city": "Paris",
        "lat": "49.32545",
        "lon": "4.2565",
    }
    index_document(doc)
    assert DB.zscore("w|vernou", "d|yyyy") == 4
    assert DB.zscore("w|celle", "d|yyyy") == 4 / 5


def test_deindex_document_should_deindex_list_values():
    doc = {
        "id": "xxxx",
        "_id": "yyyy",
        "type": "street",
        "name": ["Vernou-la-Celle-sur-Seine", "Vernou"],
        "city": "Paris",
        "lat": "49.32545",
        "lon": "4.2565",
    }
    index_document(doc)
    deindex_document(doc["_id"])
    assert not ds._DB.exists("d|yyyy")
    assert not DB.exists("w|vernou")
    assert not DB.exists("w|celle")
    assert len(DB.keys()) == 0


def test_deindex_document_should_not_fail_if_id_do_not_exist():
    deindex_document("xxxxx")


def test_should_be_possible_to_define_fields_from_config(config):
    config.FIELDS = [
        {"key": "custom"},
        {"key": "special"},
    ]
    doc = {
        "id": "xxxx",
        "_id": "yyyy",
        "lat": "49.32545",
        "lon": "4.2565",
        "custom": "rue",
        "special": "Lilas",
        "thisone": "is not indexed",
    }
    index_document(doc)
    assert ds._DB.exists("d|yyyy")
    assert DB.exists("w|lilas")
    assert DB.exists("w|rue")
    assert not DB.exists("w|indexed")


def test_should_be_possible_to_override_boost_from_config(config):
    config.FIELDS = [
        {"key": "name", "boost": 5},
        {"key": "city"},
    ]
    doc = {
        "id": "xxxx",
        "_id": "yyyy",
        "lat": "49.32545",
        "lon": "4.2565",
        "name": "Lilas",
        "city": "Cergy",
    }
    index_document(doc)
    assert ds._DB.exists("d|yyyy")
    assert DB.zscore("w|lilas", "d|yyyy") == 5
    assert DB.zscore("w|cergy", "d|yyyy") == 1


def test_should_be_possible_to_override_boost_with_callable(config):
    config.FIELDS = [
        {"key": "name", "boost": lambda doc: 5},
        {"key": "city"},
    ]
    doc = {
        "id": "xxxx",
        "_id": "yyyy",
        "lat": "49.32545",
        "lon": "4.2565",
        "name": "Lilas",
        "city": "Cergy",
    }
    index_document(doc)
    assert ds._DB.exists("d|yyyy")
    assert DB.zscore("w|lilas", "d|yyyy") == 5
    assert DB.zscore("w|cergy", "d|yyyy") == 1


def test_doc_with_null_value_should_not_be_index_if_not_allowed(config):
    config.FIELDS = [
        {"key": "name", "null": False},
        {"key": "city"},
    ]
    doc = {
        "id": "xxxx",
        "_id": "yyyy",
        "lat": "49.32545",
        "lon": "4.2565",
        "name": "",
        "city": "Cergy",
    }
    index_document(doc)
    assert not DB.exists("w|cergy")


def test_doc_with_array_of_array_value_should_not_be_index_if_not_allowed(config):
    config.FIELDS = [
        {"key": "name", "null": False},
        {"key": "city"},
    ]
    doc = {
        "id": "xxxx",
        "_id": "yyyy",
        "lat": "49.32545",
        "lon": "4.2565",
        "name": [["Lilas"]],
        "city": "Cergy",
    }
    index_document(doc)
    assert not DB.exists("w|cergy")


def test_create_edge_ngrams(config):
    config.MIN_EDGE_NGRAMS = 2
    config.INDEX_EDGE_NGRAMS = False
    doc = {
        "id": "xxxx",
        "_id": "yyyy",
        "lat": "49.32545",
        "lon": "4.2565",
        "name": "28 Lilas",  # 28 should not appear in ngrams
        "city": "Paris",
    }
    index_document(doc)
    assert not DB.exists("n|li")
    assert not DB.exists("n|lil")
    assert not DB.exists("n|lila")
    assert not DB.exists("n|pa")
    assert not DB.exists("n|par")
    assert not DB.exists("n|pari")
    create_edge_ngrams()
    assert DB.exists("n|li")
    assert DB.exists("n|lil")
    assert DB.exists("n|pa")
    assert DB.exists("n|par")
    assert not DB.exists("n|28")
    assert len(DB.keys()) == 13
    assert len(ds._DB.keys()) == 1


def test_index_document_with_custom_id(config):
    config.ID_FIELD = "custom"
    doc = DOC.copy()
    doc["custom"] = "custom_id"
    index_document(doc)
    assert ds._DB.exists("d|custom_id")



#!/usr/bin/env python

import argparse
import os

from addok import VERSION
from addok.config import config


def main():
    if VERSION:
        print("Addok", VERSION)
    main_parser = argparse.ArgumentParser(
        description="Addok command line.", add_help=False
    )
    main_parser.add_argument("--config", help="Local config")
    args, extras = main_parser.parse_known_args()
    if args.config:
        os.environ["ADDOK_CONFIG_MODULE"] = args.config
    main_parser.add_argument(
        "-h", "--help", action="store_true", help="Show this help message and exit"
    )

    subparsers = main_parser.add_subparsers(title="Available commands", metavar="")

    from addok import hooks

    config.load()
    hooks.register_command(subparsers)
    args = main_parser.parse_args(args=extras)
    if getattr(args, "func", None):
        args.func(args)
    else:
        main_parser.print_help()


if __name__ == "__main__":
    main()



from pathlib import Path

from addok.config import config
from addok.db import DB


@config.on_load
def load_scripts():
    root = Path(__file__).parent / "lua"
    for path in root.glob("*.lua"):
        with path.open() as f:
            name = path.name[:-4]
            globals()[name] = DB.register_script(f.read())



import string

from addok.db import DB
from addok.helpers import keys as dbkeys
from addok.helpers import blue, white
from addok.helpers.search import preprocess_query
from addok.helpers.text import Token
from addok.pairs import pair_key

from addok.config import config


def make_fuzzy(word, max=1):
    """Naive neighborhoods algo."""
    # inversions
    neighbors = []
    for i in range(0, len(word) - 1):
        neighbor = list(word)
        neighbor[i], neighbor[i + 1] = neighbor[i + 1], neighbor[i]
        neighbors.append("".join(neighbor))

    # limit substitutions to keyboard mapping
    if config.FUZZY_KEY_MAP is not None:
        for i in range(0, len(word)):
            neighbor = list(word)
            if neighbor[i] in config.FUZZY_KEY_MAP:
                for letter in list(config.FUZZY_KEY_MAP[neighbor[i]]):
                    if letter != neighbor[i]:
                        neighbor[i] = letter
                        neighbors.append("".join(neighbor))
    else:
        # substitutions
        for letter in string.ascii_lowercase:
            for i in range(0, len(word)):
                neighbor = list(word)
                if letter != neighbor[i]:
                    neighbor[i] = letter
                    neighbors.append("".join(neighbor))

    # insertions
    for letter in string.ascii_lowercase:
        for i in range(0, len(word) + 1):
            neighbor = list(word)
            neighbor.insert(i, letter)
            neighbors.append("".join(neighbor))

    # removal
    if len(word) > 3:
        for i in range(0, len(word)):
            neighbor = list(word)
            del neighbor[i]
            neighbors.append("".join(neighbor))

    # Order-preserving deduplication of neighbors
    neighbors = list(dict.fromkeys(neighbors))
    return neighbors


def fuzzy_collector(helper):
    if helper.fuzzy and not helper.has_cream():
        if helper.not_found:
            try_fuzzy(helper, helper.not_found)
        if helper.bucket_dry and not helper.has_cream():
            try_fuzzy(helper, helper.meaningful)
        if helper.bucket_dry and not helper.has_cream() and helper.common:
            try_fuzzy(helper, helper.meaningful, include_common=False)


def try_fuzzy(helper, tokens, include_common=True):
    if not helper.bucket_dry or not tokens:
        return
    helper.debug("Fuzzy on. Trying with %s.", tokens)
    tokens.sort(key=lambda t: len(t), reverse=True)
    allkeys = helper.keys[:]
    if include_common:
        # As we are in fuzzy, try to narrow as much as possible by adding
        # unused common tokens.
        allkeys.extend([t.db_key for t in helper.common if t.db_key not in helper.keys])
    for try_one in tokens:
        if helper.bucket_full:
            break
        keys = allkeys[:]
        if try_one.db_key in keys:
            keys.remove(try_one.db_key)
        if try_one.isdigit():
            continue
        helper.debug("Going fuzzy with %s and %s", try_one, keys)
        neighbors = make_fuzzy(try_one, max=helper.fuzzy)
        if len(keys):
            # Only retain tokens that have been seen in the index at least
            # once with the other tokens.
            DB.sadd(helper.pid, *neighbors)
            interkeys = [pair_key(k[2:]) for k in keys]
            interkeys.append(helper.pid)
            fuzzy_words = DB.sinter(interkeys)
            DB.delete(helper.pid)
            # Keep the priority we gave in building fuzzy terms (inversion
            # first, then substitution, etc.).
            fuzzy_words = [w.decode() for w in fuzzy_words]
            fuzzy_words.sort(key=lambda x: neighbors.index(x))
        else:
            # The token we are considering is alone.
            fuzzy_words = []
            for neighbor in neighbors:
                key = dbkeys.token_key(neighbor)
                count = DB.zcard(key)
                if count:
                    fuzzy_words.append(neighbor)
        if fuzzy_words:
            helper.debug("Found fuzzy candidates %s", fuzzy_words)
            fuzzy_keys = [dbkeys.token_key(w) for w in fuzzy_words]
            for key in fuzzy_keys:
                if helper.bucket_dry:
                    helper.add_to_bucket(keys + [key])


def do_fuzzy(self, word):
    """Compute fuzzy extensions of word.
    FUZZY lilas"""
    word = list(preprocess_query(word))[0]
    fuzzy = make_fuzzy(word)
    print(white(fuzzy))
    print(blue("{} items".format(len(fuzzy))))


def do_fuzzyindex(self, word):
    """Compute fuzzy extensions of word that exist in index.
    FUZZYINDEX lilas"""
    word = list(preprocess_query(word))[0]
    token = Token(word)
    neighbors = make_fuzzy(token)
    neighbors = [(n, DB.zcard(dbkeys.token_key(n))) for n in neighbors]
    neighbors.sort(key=lambda n: n[1], reverse=True)
    for token, freq in neighbors:
        if freq == 0:
            break
        print(white(token), blue(freq))


def register_shell_command(cmd):
    cmd.register_commands(do_fuzzy, do_fuzzyindex)



import geohash
import redis

from addok.config import config
from addok.db import DB
from addok.ds import get_document

from . import iter_pipe, keys, yielder

VALUE_SEPARATOR = "|~|"


def preprocess(s):
    if s not in _CACHE:
        _CACHE[s] = list(iter_pipe(s, config.PROCESSORS))
    return _CACHE[s]


_CACHE = {}


def token_key_frequency(key):
    return DB.zcard(key)


def token_frequency(token):
    return token_key_frequency(keys.token_key(token))


def extract_tokens(tokens, string, boost):
    els = list(preprocess(string))
    if not els:
        return
    boost = config.DEFAULT_BOOST / len(els) * boost
    for token in els:
        if tokens.get(token, 0) < boost:
            tokens[token] = boost


def index_tokens(pipe, tokens, key, **kwargs):
    for token, boost in tokens.items():
        pipe.zadd(keys.token_key(token), mapping={key: boost})


def deindex_field(key, string):
    els = list(preprocess(string))
    for s in els:
        deindex_token(key, s)
    return els


def deindex_token(key, token):
    tkey = keys.token_key(token)
    DB.zrem(tkey, key)


def index_documents(docs):
    pipe = DB.pipeline(transaction=False)
    for doc in docs:
        if not doc:
            continue
        if doc.get("_action") in ["delete", "update"]:
            key = keys.document_key(doc[config.ID_FIELD]).encode()
            known_doc = get_document(key)
            if known_doc:
                deindex_document(known_doc)
        if doc.get("_action") in ["index", "update", None]:
            index_document(pipe, doc)
        yield doc
    try:
        pipe.execute()
    except redis.RedisError as e:
        msg = "Error while importing document:\n{}\n{}".format(doc, str(e))
        raise ValueError(msg)


def index_document(pipe, doc, **kwargs):
    key = keys.document_key(doc[config.ID_FIELD])
    tokens = {}
    for indexer in config.INDEXERS:
        try:
            indexer.index(pipe, key, doc, tokens, **kwargs)
        except ValueError as e:
            print(e)
            return  # Do not index.


def deindex_document(doc, **kwargs):
    key = keys.document_key(doc[config.ID_FIELD])
    tokens = []
    for indexer in config.INDEXERS:
        indexer.deindex(DB, key, doc, tokens, **kwargs)


def index_geohash(pipe, key, lat, lon):
    lat = float(lat)
    lon = float(lon)
    geoh = geohash.encode(lat, lon, config.GEOHASH_PRECISION)
    geok = keys.geohash_key(geoh)
    pipe.sadd(geok, key)


def deindex_geohash(key, lat, lon):
    lat = float(lat)
    lon = float(lon)
    geoh = geohash.encode(lat, lon, config.GEOHASH_PRECISION)
    geok = keys.geohash_key(geoh)
    DB.srem(geok, key)


def check_type_and_transform_to_array(name, values):
    # Transform to array
    if isinstance(values, (float, int, str)):
        values = [values]
    # Check type
    if not all(isinstance(item, (float, int, str)) for item in values):
        raise ValueError("{} must be of type float, integer or string (or an array of these types)".format(name))
    return values


class FieldsIndexer:
    @staticmethod
    def index(pipe, key, doc, tokens, **kwargs):
        importance = float(doc.get("importance", 0.0)) * config.IMPORTANCE_WEIGHT
        for field in config.FIELDS:
            name = field["key"]
            values = doc.get(name)
            if not values:
                if not field.get("null", True):
                    # A mandatory field is null.
                    raise ValueError("{} must not be null".format(name))
                continue
            if name != config.HOUSENUMBERS_FIELD:
                boost = field.get("boost", config.DEFAULT_BOOST)
                if callable(boost):
                    boost = boost(doc)
                boost = boost + importance
                values = check_type_and_transform_to_array(name, values)
                for value in values:
                    extract_tokens(tokens, str(value), boost=boost)
        index_tokens(pipe, tokens, key, **kwargs)

    @staticmethod
    def deindex(db, key, doc, tokens, **kwargs):
        for field in config.FIELDS:
            name = field["key"]
            if name == config.HOUSENUMBERS_FIELD:
                continue
            values = doc.get(name)
            if values:
                values = check_type_and_transform_to_array(name, values)
                for value in values:
                    tokens.extend(deindex_field(key, value))


class GeohashIndexer:
    @staticmethod
    def index(pipe, key, doc, tokens, **kwargs):
        index_geohash(pipe, key, doc["lat"], doc["lon"])

    @staticmethod
    def deindex(db, key, doc, tokens, **kwargs):
        deindex_geohash(key, doc["lat"], doc["lon"])


class HousenumbersIndexer:
    @staticmethod
    def index(pipe, key, doc, tokens, **kwargs):
        housenumbers = doc.get("housenumbers", {})
        for number, data in housenumbers.items():
            index_geohash(pipe, key, data["lat"], data["lon"])

    @staticmethod
    def deindex(db, key, doc, tokens, **kwargs):
        housenumbers = doc.get("housenumbers", {})
        for token, data in housenumbers.items():
            deindex_geohash(key, data["lat"], data["lon"])


class FiltersIndexer:
    @staticmethod
    def index(pipe, key, doc, tokens, **kwargs):
        for name in config.FILTERS:
            values = doc.get(name)
            if values:
                values = check_type_and_transform_to_array(name, values)
                for value in values:
                    pipe.sadd(keys.filter_key(name, value), key)
        # Special case for housenumber type, because it's not a real type
        if (
            "type" in config.FILTERS
            and config.HOUSENUMBERS_FIELD
            and doc.get(config.HOUSENUMBERS_FIELD)
        ):
            pipe.sadd(keys.filter_key("type", "housenumber"), key)

    @staticmethod
    def deindex(db, key, doc, tokens, **kwargs):
        for name in config.FILTERS:
            values = doc.get(name)
            if values:
                values = check_type_and_transform_to_array(name, values)
                for value in values:
                    db.srem(keys.filter_key(name, value), key)
        if "type" in config.FILTERS:
            db.srem(keys.filter_key("type", "housenumber"), key)


@yielder
def prepare_housenumbers(doc):
    # We need to have the housenumbers tokenized in the document, to match
    # from user query (see results.match_housenumber).
    if not doc:
        return
    housenumbers = doc.get(config.HOUSENUMBERS_FIELD)
    if housenumbers:
        doc["housenumbers"] = {}
        for number, data in housenumbers.items():
            # Housenumber may have multiple tokens (eg.: "dix huit").
            token = "".join(list(preprocess(number)))
            data["raw"] = number
            doc["housenumbers"][token] = data
    return doc



import pytest
from addok.fuzzy import make_fuzzy
from addok.helpers.text import (
    Token,
    _normalize,
    _tokenize,
    alphanumerize,
    ascii,
    compare_str,
    compute_edge_ngrams,
    contains,
    equals,
    ngrams,
    startswith,
    synonymize,
)


@pytest.mark.parametrize(
    "input,output",
    [
        ["one two three", ["one", "two", "three"]],
        ["presqu'ile", ["presqu", "ile"]],
        ["22, rue", ["22", "rue"]],
    ],
)
def test_tokenize(input, output):
    assert _tokenize(input) == output


def test_make_fuzzy_should_extend_term(config):
    config.FUZZY_KEY_MAP = None
    expected = set(
        [
            "omt",
            "mto",
            "amot",
            "maot",
            "moat",
            "mota",
            "bmot",
            "mbot",
            "mobt",
            "motb",
            "cmot",
            "mcot",
            "moct",
            "motc",
            "dmot",
            "mdot",
            "modt",
            "motd",
            "emot",
            "meot",
            "moet",
            "mote",
            "fmot",
            "mfot",
            "moft",
            "motf",
            "gmot",
            "mgot",
            "mogt",
            "motg",
            "hmot",
            "mhot",
            "moht",
            "moth",
            "imot",
            "miot",
            "moit",
            "moti",
            "jmot",
            "mjot",
            "mojt",
            "motj",
            "kmot",
            "mkot",
            "mokt",
            "motk",
            "lmot",
            "mlot",
            "molt",
            "motl",
            "mmot",
            "mmot",
            "momt",
            "motm",
            "nmot",
            "mnot",
            "mont",
            "motn",
            "omot",
            "moot",
            "moot",
            "moto",
            "pmot",
            "mpot",
            "mopt",
            "motp",
            "qmot",
            "mqot",
            "moqt",
            "motq",
            "rmot",
            "mrot",
            "mort",
            "motr",
            "smot",
            "msot",
            "most",
            "mots",
            "tmot",
            "mtot",
            "mott",
            "mott",
            "umot",
            "muot",
            "mout",
            "motu",
            "vmot",
            "mvot",
            "movt",
            "motv",
            "wmot",
            "mwot",
            "mowt",
            "motw",
            "xmot",
            "mxot",
            "moxt",
            "motx",
            "ymot",
            "myot",
            "moyt",
            "moty",
            "zmot",
            "mzot",
            "mozt",
            "motz",
            "aot",
            "mat",
            "moa",
            "bot",
            "mbt",
            "mob",
            "cot",
            "mct",
            "moc",
            "dot",
            "mdt",
            "mod",
            "eot",
            "met",
            "moe",
            "fot",
            "mft",
            "mof",
            "got",
            "mgt",
            "mog",
            "hot",
            "mht",
            "moh",
            "iot",
            "mit",
            "moi",
            "jot",
            "mjt",
            "moj",
            "kot",
            "mkt",
            "mok",
            "lot",
            "mlt",
            "mol",
            "mmt",
            "mom",
            "not",
            "mnt",
            "mon",
            "oot",
            "moo",
            "pot",
            "mpt",
            "mop",
            "qot",
            "mqt",
            "moq",
            "rot",
            "mrt",
            "mor",
            "sot",
            "mst",
            "mos",
            "tot",
            "mtt",
            "uot",
            "mut",
            "mou",
            "vot",
            "mvt",
            "mov",
            "wot",
            "mwt",
            "mow",
            "xot",
            "mxt",
            "mox",
            "yot",
            "myt",
            "moy",
            "zot",
            "mzt",
            "moz",
        ]
    )
    assert set(make_fuzzy("mot")) == expected


def test_make_fuzzy_with_key_map_should_extend_term():
    expected = set(
        [
            "omt",
            "mto",
            "lot",
            "pot",
            "uot",
            "mit",
            "mat",
            "mkt",
            "mlt",
            "mpt",
            "mor",
            "mof",
            "mog",
            "moy",
            "amot",
            "maot",
            "moat",
            "mota",
            "bmot",
            "mbot",
            "mobt",
            "motb",
            "cmot",
            "mcot",
            "moct",
            "motc",
            "dmot",
            "mdot",
            "modt",
            "motd",
            "emot",
            "meot",
            "moet",
            "mote",
            "fmot",
            "mfot",
            "moft",
            "motf",
            "gmot",
            "mgot",
            "mogt",
            "motg",
            "hmot",
            "mhot",
            "moht",
            "moth",
            "imot",
            "miot",
            "moit",
            "moti",
            "jmot",
            "mjot",
            "mojt",
            "motj",
            "kmot",
            "mkot",
            "mokt",
            "motk",
            "lmot",
            "mlot",
            "molt",
            "motl",
            "mmot",
            "momt",
            "motm",
            "nmot",
            "mnot",
            "mont",
            "motn",
            "omot",
            "moot",
            "moto",
            "pmot",
            "mpot",
            "mopt",
            "motp",
            "qmot",
            "mqot",
            "moqt",
            "motq",
            "rmot",
            "mrot",
            "mort",
            "motr",
            "smot",
            "msot",
            "most",
            "mots",
            "tmot",
            "mtot",
            "mott",
            "umot",
            "muot",
            "mout",
            "motu",
            "vmot",
            "mvot",
            "movt",
            "motv",
            "wmot",
            "mwot",
            "mowt",
            "motw",
            "xmot",
            "mxot",
            "moxt",
            "motx",
            "ymot",
            "myot",
            "moyt",
            "moty",
            "zmot",
            "mzot",
            "mozt",
            "motz",
        ]
    )
    assert set(make_fuzzy("mot")) == expected


def test_make_fuzzy_should_remove_letter_if_world_is_long():
    assert "mt" not in make_fuzzy("mot")
    assert "rain" in make_fuzzy("train")
    assert "tain" in make_fuzzy("train")
    assert "trin" in make_fuzzy("train")
    assert "tran" in make_fuzzy("train")
    assert "trai" in make_fuzzy("train")


@pytest.mark.parametrize(
    "left,right,score",
    [
        ["Lille", "Lille", 1],
        ["Lille", "lille", 1],
        ["Andrésy", "andresy", 1],
        ["Y", "y", 1],
        ["Ay", "ay", 1],
    ],
)
def test_compare_str(left, right, score):
    assert compare_str(left, right) == score


@pytest.mark.parametrize(
    "best,other,query",
    [
        [
            "avenue de paris 94123 saint mande",
            "avenue de saint mande 75012 paris",
            "avenue de paris saint mande",
        ],
        [
            "1 place du trocadero et du 11 novembre 75016 paris",
            "square du trocadero 75016 paris",
            "place du trocadero paris",
        ],
    ],
)
def test_compare_strs(best, other, query):
    assert compare_str(best, query) > compare_str(other, query)


@pytest.mark.parametrize(
    "input,n,output",
    [
        ["Lille", 2, {" L", "Li", "il", "ll", "le", "e "}],
        ["Lille", 3, {" Li", "Lil", "ill", "lle", "le "}],
        ["L", 3, {" L "}],
    ],
)
def test_ngrams(input, n, output):
    assert ngrams(input, n) == output


@pytest.mark.parametrize(
    "input,output",
    [
        ["ABCDEF", "abcdef"],
        ["éêàù", "eeau"],
        ["Étretat", "etretat"],
        ["Erispœ", "erispoe"],
    ],
)
def test_normalize(input, output):
    assert _normalize(Token(input)) == output


@pytest.mark.parametrize(
    "input,output",
    [
        ["rue d'Andrésy", "rue d Andrésy"],
        ["a   ", "a"],
        ["(défibrillateur)", "défibrillateur"],
        ["machin (défibrillateur)", "machin défibrillateur"],
    ],
)
def test_alphanumerize(input, output):
    assert alphanumerize(Token(input)) == output


@pytest.mark.parametrize(
    "input,output",
    [
        ["bd", ["boulevard"]],
        ["13e", ["treizieme"]],
        ["18e", ["dix", "huitieme"]],
    ],
)
def test_synonymize(input, output, config):
    # Make sure we control synonyms.
    config.SYNONYMS = {"bd": "boulevard", "13e": "treizieme", "18e": "dix huitieme"}
    assert list(synonymize([Token(input)])) == output


def test_synonyms_file_is_loaded(config):
    # See synonyms.txt
    assert config.SYNONYMS == {
        "cba": "abc",
        "xzy": "xyz",
        "zyx": "xyz",
    }


def test_compute_edge_ngrams():
    assert compute_edge_ngrams("vanbrechi") == [
        "van",
        "vanb",
        "vanbr",
        "vanbre",
        "vanbrec",
        "vanbrech",
    ]


def test_compute_edge_ngrams_honor_min_edge_ngrams_setting(config):
    config.MIN_EDGE_NGRAMS = 1
    assert compute_edge_ngrams("abcd") == ["a", "ab", "abc"]


def test_compute_edge_ngrams_honor_max_edge_ngrams_setting(config):
    config.MAX_EDGE_NGRAMS = 5
    assert compute_edge_ngrams("abcdefghijklmn") == ["abc", "abcd", "abcde"]


@pytest.mark.parametrize(
    "candidate,target",
    [
        ["22 rue vicq", "22 Rue Vicq d'Azir 75010 Paris"],
        ["rue vicq", "22 Rue Vicq d'Azir 75010 Paris"],
    ],
)
def test_contains(candidate, target):
    assert contains(candidate, target)


@pytest.mark.parametrize(
    "candidate,target",
    [
        ["22 rue vicq", "22 Rue Vicq d'Azir 75010 Paris"],
        ["etang des rivieres", "Étang des Rivières 42330 Saint-Galmier"],
    ],
)
def test_startswith(candidate, target):
    assert startswith(candidate, target)


@pytest.mark.parametrize(
    "candidate,target",
    [
        ["22 rue vicq d azir 75010 paris", "22 Rue Vicq d'Azir 75010 Paris"],
        ["etang des rivieres", "Étang des Rivières"],
        ["Saint galmier", "Saint-Galmier"],
    ],
)
def test_equals(candidate, target):
    assert equals(candidate, target)


def test_ascii_should_behave_like_a_string():
    s = ascii("mystring")
    assert str(s) == "mystring"


def test_ascii_should_clean_string():
    s = ascii("Aystringé")
    assert s == "aystringe"


def test_ascii_should_cache_cleaned_string(monkeypatch):
    s = ascii("mystring")
    assert s._cache

    def do_not_call_me(x):
        assert False


def test_ngrams_should_cache():
    ngrams("test")
    assert ngrams.cache_info() is not None



from addok.db import DB
from addok.helpers import keys, magenta, white
from addok.helpers.search import preprocess_query


def pair_key(s):
    return "p|{}".format(s)


class PairsIndexer:
    @staticmethod
    def index(pipe, key, doc, tokens, **kwargs):
        for token in list(set(tokens.keys())):  # Unique values.
            pairs = set(t for t in tokens if t != token)
            if pairs:
                pipe.sadd(pair_key(token), *pairs)

    @staticmethod
    def deindex(db, key, doc, tokens, **kwargs):
        tokens = list(set(tokens))  # Unique values.
        for i, token in enumerate(tokens):
            for token2 in tokens[i:]:
                if token != token2:
                    tmp_key = "|".join(["didx", token, token2])
                    # Do we have other documents that share token and token2?
                    commons = db.zinterstore(
                        tmp_key, [keys.token_key(token), keys.token_key(token2)]
                    )
                    db.delete(tmp_key)
                    if not commons:
                        db.srem(pair_key(token), token2)
                        db.srem(pair_key(token2), token)


def pair(cmd, word):
    """See all token associated with a given token.
    PAIR lilas"""
    word = list(preprocess_query(word))[0]
    key = pair_key(word)
    tokens = [t.decode() for t in DB.smembers(key)]
    tokens.sort()
    print(white(tokens))
    print(magenta("(Total: {})".format(len(tokens))))


def register_shell_command(cmd):
    cmd.register_command(pair)



from addok.config import config
from addok.helpers import haversine_distance, km_to_score
from addok.helpers.text import (
    ascii,
    compare_ngrams,
    compare_str,
    contains,
    equals,
    startswith,
)


def make_labels(helper, result):
    if not result.labels:
        # Make your own for better scoring (see addok-france for inspiration).
        result.labels = result._rawattr(config.NAME_FIELD)[:]
        label = result.labels[0]
        city = getattr(result, "city", None)
        if city and city != label:
            postcode = getattr(result, "postcode", None)
            if postcode:
                label = label + " " + postcode
            label = label + " " + city
        housenumber = getattr(result, "housenumber", None)
        if housenumber:
            label = "{} {}".format(housenumber, label)
        # Replace default label with our computed one, but keep the other raw
        # aliases, and let plugins add more of them if needed.
        result.labels[0] = label


def match_housenumber(helper, result):
    if not helper.check_housenumber:
        return
    # Housenumber may have multiple tokens (eg. "dix huit"), we join
    # those to match the way they have been processed by
    # addok.helpers.index.prepare_housenumbers.
    raw = "".join(sorted(helper.housenumbers, key=lambda t: t.position))
    if raw and raw in result.housenumbers:
        data = result.housenumbers[str(raw)]
        result.housenumber = data.pop("raw")
        result.type = "housenumber"
        result.update(data)
    if helper.only_housenumber and not result.housenumber:
        helper.debug(
            "Cannot match housenumber (%s), removing `%s`", result.housenumber, result
        )
        return False


def score_by_importance(helper, result):
    importance = getattr(result, "importance", None) or 0.0
    result.add_score(
        "importance",
        float(importance) * config.IMPORTANCE_WEIGHT,
        config.IMPORTANCE_WEIGHT,
    )


def score_by_autocomplete_distance(helper, result):
    if not helper.autocomplete:
        return
    score = 0
    query = ascii(helper.query)
    for idx, label in enumerate(result.labels):
        label = ascii(label)
        result.labels[idx] = label  # Cache ascii folding.
        if equals(query, label):
            score = 1.0
        elif startswith(query, label):
            score = 0.9
        elif contains(query, label):
            score = 0.7
        if score:
            result.add_score("str_distance", score, ceiling=1.0)
    if not score:
        _score_by_ngram_distance(helper, result, scale=0.9)


def _score_by_str_distance(helper, result, scale=1.0):
    for label in result.labels:
        score = compare_str(label, helper.query) * scale
        result.add_score("str_distance", score, ceiling=1.0)


def score_by_str_distance(helper, result):
    if helper.autocomplete:
        return
    _score_by_str_distance(helper, result)


def _score_by_ngram_distance(helper, result, scale=1.0):
    for label in result.labels:
        label = ascii(label)
        score = compare_ngrams(label, helper.query) * scale
        result.add_score("str_distance", score, ceiling=1.0)
        if score >= config.MATCH_THRESHOLD:
            break


def score_by_ngram_distance(helper, result):
    if helper.autocomplete:
        return
    _score_by_ngram_distance(helper, result)


def score_by_geo_distance(helper, result):
    if helper.lat is None or helper.lon is None:
        return
    km = haversine_distance(
        (float(result.lat), float(result.lon)), (helper.lat, helper.lon)
    )
    result.distance = km * 1000
    result.add_score(
        "geo_distance",
        km_to_score(km) * config.GEO_DISTANCE_WEIGHT,
        ceiling=config.GEO_DISTANCE_WEIGHT,
    )


def adjust_scores(helper, result):
    if helper.lat is not None and helper.lon is not None:
        str_distance = result._scores.get("str_distance")
        if str_distance:
            result._scores["str_distance"] = (str_distance[0] * 0.9, str_distance[1])
        importance = result._scores.get("importance")
        if importance:
            result._scores["importance"] = (importance[0] * 0.1, importance[1])


def load_closer(helper, result):
    if not helper.check_housenumber:
        return

    def sort(h):
        return haversine_distance(
            (float(h["lat"]), float(h["lon"])), (helper.lat, helper.lon)
        )

    candidates = []
    if result.housenumbers:
        candidates = list(result.housenumbers.values())
    if not helper.only_housenumber:
        candidates.append({"raw": None, "lat": result.lat, "lon": result.lon})
    candidates.sort(key=sort)
    closer = candidates[0]
    if closer["raw"]:  # Means a housenumber is closer than street center.
        result.housenumber = closer.pop("raw")
        result.type = "housenumber"
        result.update(closer)
    if helper.only_housenumber and not result.housenumber:
        helper.debug("Removing non housenumber match `%s`", result)
        return False
